@@ -1,6 +1,6 @@
 /*  test/hfile.c -- Test cases for low-level input/output streams.
 
-    Copyright (C) 2013-2014, 2016 Genome Research Ltd.
+    Copyright (C) 2013-2014, 2016, 2018 Genome Research Ltd.
 
     Author: John Marshall <jm18@sanger.ac.uk>
 
@@ -33,6 +33,7 @@ DEALINGS IN THE SOFTWARE.  */
 
 #include "htslib/hfile.h"
 #include "htslib/hts_defs.h"
+#include "htslib/kstring.h"
 
 void HTS_NORETURN fail(const char *format, ...)
 {
@@ -202,6 +203,47 @@ int main(void)
     if ((c = hgetc(fin)) != EOF) fail("chars: hgetc (EOF) returned %d", c);
     if (hclose(fin) != 0) fail("hclose(test/hfile_chars.tmp) for reading");
 
+    fin = hopen("preload:test/hfile_chars.tmp", "r");
+    if (fin == NULL) fail("preloading \"test/hfile_chars.tmp\" for reading");
+    for (i = 0; i < 256; i++)
+        if ((c = hgetc(fin)) != i)
+            fail("preloading chars: hgetc (%d = 0x%x) returned %d = 0x%x", i, i, c, c);
+    if ((c = hgetc(fin)) != EOF) fail("preloading chars: hgetc (EOF) returned %d", c);
+    if (hclose(fin) != 0) fail("preloading hclose(test/hfile_chars.tmp) for reading");
+
+    char* test_string = strdup("Test string");
+    fin = hopen("mem:", "r:", test_string, 12);
+    if (fin == NULL) fail("hopen(\"mem:\", \"r:\", ...)");
+    if (hread(fin, buffer, 12) != 12)
+        fail("hopen('mem:', 'r') failed read");
+    if(strcmp(buffer, test_string) != 0)
+        fail("hopen('mem:', 'r') missread '%s' != '%s'", buffer, test_string);
+    char* internal_buf;
+    size_t interval_buf_len;
+    if((internal_buf = hfile_mem_get_buffer(fin, &interval_buf_len)) == NULL){
+        fail("hopen('mem:', 'r') failed to get internal buffer");
+    }
+    if (hclose(fin) != 0) fail("hclose mem for reading");
+
+    test_string = strdup("Test string");
+    fin = hopen("mem:", "wr:", test_string, 12);
+    if (fin == NULL) fail("hopen(\"mem:\", \"w:\", ...)");
+    if (hseek(fin, -1, SEEK_END) < 0)
+        fail("hopen('mem:', 'wr') failed seek");
+    if (hwrite(fin, " extra", 7) != 7)
+        fail("hopen('mem:', 'wr') failed write");
+    if (hseek(fin, 0, SEEK_SET) < 0)
+        fail("hopen('mem:', 'wr') failed seek");
+    if (hread(fin, buffer, 18) != 18)
+        fail("hopen('mem:', 'wr') failed read");
+    if (strcmp(buffer, "Test string extra") != 0)
+        fail("hopen('mem:', 'wr') misswrote '%s' != '%s'", buffer, "Test string extra");
+    if((internal_buf = hfile_mem_steal_buffer(fin, &interval_buf_len)) == NULL){
+        fail("hopen('mem:', 'wr') failed to get internal buffer");
+    }
+    free(internal_buf);
+    if (hclose(fin) != 0) fail("hclose mem for writing");
+
     fin = hopen("data:,hello, world!%0A", "r");
     if (fin == NULL) fail("hopen(\"data:...\")");
     n = hread(fin, buffer, 300);
@@ -210,10 +252,10 @@ int main(void)
     if (strcmp(buffer, "hello, world!\x0A") != 0) fail("hread result");
     if (hclose(fin) != 0) fail("hclose(\"data:...\")");
 
-    fin = hopen("test/xx#blank.sam", "r");
-    if (fin == NULL) fail("hopen(\"test/xx#blank.sam\") for reading");
-    if (hread(fin, buffer, 100) != 0) fail("test/xx#blank.sam is non-empty");
-    if (hclose(fin) != 0) fail("hclose(\"test/xx#blank.sam\") for reading");
+    fin = hopen("test/emptyfile", "r");
+    if (fin == NULL) fail("hopen(\"test/emptyfile\") for reading");
+    if (hread(fin, buffer, 100) != 0) fail("test/emptyfile is non-empty");
+    if (hclose(fin) != 0) fail("hclose(\"test/emptyfile\") for reading");
 
     fin = hopen("data:,", "r");
     if (fin == NULL) fail("hopen(\"data:\") for reading");
@@ -238,5 +280,30 @@ int main(void)
         fail("hread result for base64");
     if (hclose(fin) != 0) fail("hclose(\"data:;base64,...\")");
 
+    kstring_t kstr = { 0, 0, NULL };
+
+    if (strcmp(haddextension(&kstr, "foo/bar.bam", 0, ".bai"),
+               "foo/bar.bam.bai") != 0) fail("haddextension foo/bar.bam[.bai]");
+    if (strcmp(haddextension(&kstr, "foo/bar.bam", 1, ".bai"),
+               "foo/bar.bai") != 0) fail("haddextension foo/bar[.bai]");
+    if (strcmp(haddextension(&kstr, "foo.bar/baz", 1, ".bai"),
+               "foo.bar/baz.bai") != 0) fail("haddextension foo.bar/baz[.bai]");
+    if (strcmp(haddextension(&kstr, "foo#bar.bam", 0, ".bai"),
+               "foo#bar.bam.bai") != 0) fail("haddextension foo#bar.bam[.bai]");
+    if (strcmp(haddextension(&kstr, ".bam", 1, ".bai"),
+               ".bai") != 0) fail("haddextension [.bai]");
+    if (strcmp(haddextension(&kstr, "foo", 1, ".csi"),
+               "foo.csi") != 0) fail("haddextension foo[.csi]");
+
+    if (strcmp(haddextension(&kstr, "http://host/bar.cram?a&b&c", 0, ".crai"),
+               "http://host/bar.cram.crai?a&b&c") != 0)
+        fail("haddextension http://host/bar.cram[.crai]?a&b&c");
+
+    if (strcmp(haddextension(&kstr, "http://host/bar.cram#frag", 1, ".crai"),
+               "http://host/bar.crai#frag") != 0)
+        fail("haddextension http://host/bar[.crai]#frag");
+
+    free(ks_release(&kstr));
+
     return EXIT_SUCCESS;
 }
