@@ -1,6 +1,6 @@
 /*  thread_pool.c -- A pool of generic worker threads
 
-    Copyright (c) 2013-2017 Genome Research Ltd.
+    Copyright (c) 2013-2019 Genome Research Ltd.
 
     Author: James Bonfield <jkb@sanger.ac.uk>
 
@@ -23,6 +23,7 @@ FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 DEALINGS IN THE SOFTWARE.  */
 
 #ifndef TEST_MAIN
+#define HTS_BUILDING_LIBRARY // Enables HTSLIB_EXPORT, see htslib/hts_defs.h
 #include <config.h>
 #endif
 
@@ -40,6 +41,9 @@ DEALINGS IN THE SOFTWARE.  */
 
 #include "thread_pool_internal.h"
 
+static void hts_tpool_process_detach_locked(hts_tpool *p,
+                                            hts_tpool_process *q);
+
 //#define DEBUG
 
 #ifdef DEBUG
@@ -96,11 +100,15 @@ static int hts_tpool_add_result(hts_tpool_job *j, void *data) {
         return 0;
     }
 
-    if (!(r = malloc(sizeof(*r))))
+    if (!(r = malloc(sizeof(*r)))) {
+        pthread_mutex_unlock(&q->p->pool_m);
+        hts_tpool_process_shutdown(q);
         return -1;
+    }
 
     r->next = NULL;
     r->data = data;
+    r->result_cleanup = j->result_cleanup;
     r->serial = j->serial;
 
     q->n_output++;
@@ -111,10 +119,14 @@ static int hts_tpool_add_result(hts_tpool_job *j, void *data) {
         q->output_head = q->output_tail = r;
     }
 
-    DBG_OUT(stderr, "%d: Broadcasting result_avail (id %"PRId64")\n",
-            worker_id(j->p), r->serial);
-    pthread_cond_broadcast(&q->output_avail_c);
-    DBG_OUT(stderr, "%d: Broadcast complete\n", worker_id(j->p));
+    assert(r->serial >= q->next_serial    // Or it will never be dequeued ...
+           || q->next_serial == INT_MAX); // ... unless flush in progress.
+    if (r->serial == q->next_serial) {
+        DBG_OUT(stderr, "%d: Broadcasting result_avail (id %"PRId64")\n",
+                worker_id(j->p), r->serial);
+        pthread_cond_broadcast(&q->output_avail_c);
+        DBG_OUT(stderr, "%d: Broadcast complete\n", worker_id(j->p));
+    }
 
     pthread_mutex_unlock(&q->p->pool_m);
 
@@ -155,7 +167,8 @@ static hts_tpool_result *hts_tpool_next_result_locked(hts_tpool_process *q) {
             // Not technically input full, but can guarantee there is
             // room for the input to go somewhere so we still signal.
             // The waiting code will then check the condition again.
-            pthread_cond_signal(&q->input_not_full_c);
+            if (q->n_input < q->qsize)
+                pthread_cond_signal(&q->input_not_full_c);
             if (!q->shutdown)
                 wake_next_worker(q, 1);
         }
@@ -293,13 +306,17 @@ int hts_tpool_process_sz(hts_tpool_process *q) {
  * This sets the shutdown flag and wakes any threads waiting on process
  * condition variables.
  */
-void hts_tpool_process_shutdown(hts_tpool_process *q) {
-    pthread_mutex_lock(&q->p->pool_m);
+static void hts_tpool_process_shutdown_locked(hts_tpool_process *q) {
     q->shutdown = 1;
     pthread_cond_broadcast(&q->output_avail_c);
     pthread_cond_broadcast(&q->input_not_full_c);
     pthread_cond_broadcast(&q->input_empty_c);
     pthread_cond_broadcast(&q->none_processing_c);
+}
+
+void hts_tpool_process_shutdown(hts_tpool_process *q) {
+    pthread_mutex_lock(&q->p->pool_m);
+    hts_tpool_process_shutdown_locked(q);
     pthread_mutex_unlock(&q->p->pool_m);
 }
 
@@ -350,6 +367,7 @@ hts_tpool_process *hts_tpool_process_init(hts_tpool *p, int qsize, int in_only)
     q->output_tail = NULL;
     q->next_serial = 0;
     q->curr_serial = 0;
+    q->no_more_input = 0;
     q->n_input     = 0;
     q->n_output    = 0;
     q->n_processing= 0;
@@ -376,11 +394,19 @@ void hts_tpool_process_destroy(hts_tpool_process *q) {
     if (!q)
         return;
 
+    // Prevent dispatch from queuing up any more jobs.
+    // We want to reset (and flush) the queue here, before
+    // we set the shutdown flag, but we need to avoid races
+    // with queue more input during reset.
+    pthread_mutex_lock(&q->p->pool_m);
+    q->no_more_input = 1;
+    pthread_mutex_unlock(&q->p->pool_m);
+
     // Ensure it's fully drained before destroying the queue
     hts_tpool_process_reset(q, 0);
     pthread_mutex_lock(&q->p->pool_m);
-    hts_tpool_process_detach(q->p, q);
-    hts_tpool_process_shutdown(q);
+    hts_tpool_process_detach_locked(q->p, q);
+    hts_tpool_process_shutdown_locked(q);
 
     // Maybe a worker is scanning this queue, so delay destruction
     if (--q->ref_count > 0) {
@@ -424,10 +450,10 @@ void hts_tpool_process_attach(hts_tpool *p, hts_tpool_process *q) {
     pthread_mutex_unlock(&p->pool_m);
 }
 
-void hts_tpool_process_detach(hts_tpool *p, hts_tpool_process *q) {
-    pthread_mutex_lock(&p->pool_m);
+static void hts_tpool_process_detach_locked(hts_tpool *p,
+                                            hts_tpool_process *q) {
     if (!p->q_head || !q->prev || !q->next)
-        goto done;
+        return;
 
     hts_tpool_process *curr = p->q_head, *first = curr;
     do {
@@ -445,8 +471,11 @@ void hts_tpool_process_detach(hts_tpool *p, hts_tpool_process *q) {
 
         curr = curr->next;
     } while (curr != first);
+}
 
- done:
+void hts_tpool_process_detach(hts_tpool *p, hts_tpool_process *q) {
+    pthread_mutex_lock(&p->pool_m);
+    hts_tpool_process_detach_locked(p, q);
     pthread_mutex_unlock(&p->pool_m);
 }
 
@@ -472,18 +501,15 @@ static void *tpool_worker(void *arg) {
     hts_tpool *p = w->p;
     hts_tpool_job *j;
 
-    for (;;) {
+    pthread_mutex_lock(&p->pool_m);
+    while (!p->shutdown) {
         // Pop an item off the pool queue
-        pthread_mutex_lock(&p->pool_m);
 
         assert(p->q_head == 0 || (p->q_head->prev && p->q_head->next));
 
         int work_to_do = 0;
         hts_tpool_process *first = p->q_head, *q = first;
         do {
-            if (p->shutdown)
-                break;
-
             // Iterate over queues, finding one with jobs and also
             // room to put the result.
             //if (q && q->input_head && !hts_tpool_process_output_full(q)) {
@@ -496,15 +522,6 @@ static void *tpool_worker(void *arg) {
             if (q) q = q->next;
         } while (q && q != first);
 
-        if (p->shutdown) {
-        shutdown:
-#ifdef DEBUG
-            fprintf(stderr, "%d: Shutting down\n", worker_id(p));
-#endif
-            pthread_mutex_unlock(&p->pool_m);
-            return NULL;
-        }
-
         if (!work_to_do) {
             // We scanned all queues and cannot process any, so we wait.
             p->nwaiting++;
@@ -531,8 +548,7 @@ static void *tpool_worker(void *arg) {
             }
 
             p->nwaiting--;
-            pthread_mutex_unlock(&p->pool_m);
-            continue; // To outer for(;;) loop.
+            continue; // To outer loop.
         }
 
         // Otherwise work_to_do, so process as many items in this queue as
@@ -569,21 +585,35 @@ static void *tpool_worker(void *arg) {
             DBG_OUT(stderr, "%d: Processing queue %p, serial %"PRId64"\n",
                     worker_id(j->p), q, j->serial);
 
-            hts_tpool_add_result(j, j->func(j->arg));
+            if (hts_tpool_add_result(j, j->func(j->arg)) < 0)
+                goto err;
             //memset(j, 0xbb, sizeof(*j));
             free(j);
 
             pthread_mutex_lock(&p->pool_m);
         }
-        if (--q->ref_count == 0) // we were the last user
+        if (--q->ref_count == 0) { // we were the last user
             hts_tpool_process_destroy(q);
-        else
+        } else {
             // Out of jobs on this queue, so restart search from next one.
             // This is equivalent to "work-stealing".
-            p->q_head = q->next;
-
-        pthread_mutex_unlock(&p->pool_m);
+            if (p->q_head)
+                p->q_head = p->q_head->next;
+        }
     }
+
+ shutdown:
+    pthread_mutex_unlock(&p->pool_m);
+#ifdef DEBUG
+    fprintf(stderr, "%d: Shutting down\n", worker_id(p));
+#endif
+    return NULL;
+
+ err:
+#ifdef DEBUG
+    fprintf(stderr, "%d: Failed to add result\n", worker_id(p));
+#endif
+    return NULL;
 }
 
 static void wake_next_worker(hts_tpool_process *q, int locked) {
@@ -713,7 +743,7 @@ int hts_tpool_size(hts_tpool *p) {
  */
 int hts_tpool_dispatch(hts_tpool *p, hts_tpool_process *q,
                     void *(*func)(void *arg), void *arg) {
-    return hts_tpool_dispatch2(p, q, func, arg, 0);
+    return hts_tpool_dispatch3(p, q, func, arg, NULL, NULL, 0);
 }
 
 /*
@@ -724,7 +754,15 @@ int hts_tpool_dispatch(hts_tpool *p, hts_tpool_process *q,
  * nonblock -1 => add task regardless of whether queue is full (over-size)
  */
 int hts_tpool_dispatch2(hts_tpool *p, hts_tpool_process *q,
-                     void *(*func)(void *arg), void *arg, int nonblock) {
+                        void *(*func)(void *arg), void *arg, int nonblock) {
+    return hts_tpool_dispatch3(p, q, func, arg, NULL, NULL, nonblock);
+}
+
+int hts_tpool_dispatch3(hts_tpool *p, hts_tpool_process *q,
+                        void *(*exec_func)(void *arg), void *arg,
+                        void (*job_cleanup)(void *arg),
+                        void (*result_cleanup)(void *data),
+                        int nonblock) {
     hts_tpool_job *j;
 
     pthread_mutex_lock(&p->pool_m);
@@ -732,7 +770,7 @@ int hts_tpool_dispatch2(hts_tpool *p, hts_tpool_process *q,
     DBG_OUT(stderr, "Dispatching job for queue %p, serial %"PRId64"\n",
             q, q->curr_serial);
 
-    if (q->n_input >= q->qsize && nonblock == 1) {
+    if ((q->no_more_input || q->n_input >= q->qsize) && nonblock == 1) {
         pthread_mutex_unlock(&p->pool_m);
         errno = EAGAIN;
         return -1;
@@ -742,17 +780,21 @@ int hts_tpool_dispatch2(hts_tpool *p, hts_tpool_process *q,
         pthread_mutex_unlock(&p->pool_m);
         return -1;
     }
-    j->func = func;
+    j->func = exec_func;
     j->arg = arg;
+    j->job_cleanup = job_cleanup;
+    j->result_cleanup = result_cleanup;
     j->next = NULL;
     j->p = p;
     j->q = q;
     j->serial = q->curr_serial++;
 
     if (nonblock == 0) {
-        while (q->n_input >= q->qsize && !q->shutdown && !q->wake_dispatch)
+        while ((q->no_more_input || q->n_input >= q->qsize) &&
+               !q->shutdown && !q->wake_dispatch) {
             pthread_cond_wait(&q->input_not_full_c, &q->p->pool_m);
-        if (q->shutdown) {
+        }
+        if (q->no_more_input || q->shutdown) {
             free(j);
             pthread_mutex_unlock(&p->pool_m);
             return -1;
@@ -812,39 +854,39 @@ void hts_tpool_wake_dispatch(hts_tpool_process *q) {
  *        -1 on failure
  */
 int hts_tpool_process_flush(hts_tpool_process *q) {
-     int i;
-     hts_tpool *p = q->p;
+    int i;
+    hts_tpool *p = q->p;
 
-     DBG_OUT(stderr, "Flushing pool %p\n", p);
+    DBG_OUT(stderr, "Flushing pool %p\n", p);
 
-     // Drains the queue
-     pthread_mutex_lock(&p->pool_m);
+    // Drains the queue
+    pthread_mutex_lock(&p->pool_m);
 
-     // Wake up everything for the final sprint!
-     for (i = 0; i < p->tsize; i++)
-         if (p->t_stack[i])
-             pthread_cond_signal(&p->t[i].pending_c);
+    // Wake up everything for the final sprint!
+    for (i = 0; i < p->tsize; i++)
+        if (p->t_stack[i])
+            pthread_cond_signal(&p->t[i].pending_c);
 
-     // Ensure there is room for the final sprint.
-     // Shouldn't be possible to get here, but just incase.
-     if (q->qsize < q->n_output + q->n_input + q->n_processing)
-         q->qsize = q->n_output + q->n_input + q->n_processing;
+    // Ensure there is room for the final sprint.
+    // Shouldn't be possible to get here, but just incase.
+    if (q->qsize < q->n_output + q->n_input + q->n_processing)
+        q->qsize = q->n_output + q->n_input + q->n_processing;
 
-     // Wait for n_input and n_processing to hit zero.
-     while (q->n_input || q->n_processing) {
-         while (q->n_input)
-             pthread_cond_wait(&q->input_empty_c, &p->pool_m);
-         if (q->shutdown) break;
-         while (q->n_processing)
-             pthread_cond_wait(&q->none_processing_c, &p->pool_m);
-         if (q->shutdown) break;
+    // Wait for n_input and n_processing to hit zero.
+    while (q->n_input || q->n_processing) {
+        while (q->n_input)
+            pthread_cond_wait(&q->input_empty_c, &p->pool_m);
+        if (q->shutdown) break;
+        while (q->n_processing)
+            pthread_cond_wait(&q->none_processing_c, &p->pool_m);
+        if (q->shutdown) break;
     }
 
-     pthread_mutex_unlock(&p->pool_m);
+    pthread_mutex_unlock(&p->pool_m);
 
-     DBG_OUT(stderr, "Flushed complete for pool %p, queue %p\n", p, q);
+    DBG_OUT(stderr, "Flushed complete for pool %p, queue %p\n", p, q);
 
-     return 0;
+    return 0;
 }
 
 /*
@@ -860,44 +902,52 @@ int hts_tpool_process_flush(hts_tpool_process *q) {
  *        -1 on failure
  */
 int hts_tpool_process_reset(hts_tpool_process *q, int free_results) {
+    hts_tpool_job *j, *jn, *j_head;
+    hts_tpool_result *r, *rn, *r_head;
+
     pthread_mutex_lock(&q->p->pool_m);
     // prevent next_result from returning data during our flush
     q->next_serial = INT_MAX;
 
-    // Purge any queued input not yet being acted upon
-    hts_tpool_job *j, *jn;
-    for (j = q->input_head; j; j = jn) {
-        //fprintf(stderr, "Discard input %d\n", j->serial);
-        jn = j->next;
-        free(j);
-    }
+    // Remove any queued input not yet being acted upon
+    j_head = q->input_head;
     q->input_head = q->input_tail = NULL;
     q->n_input = 0;
 
-    // Purge any queued output, thus ensuring we have room to flush.
-    hts_tpool_result *r, *rn;
-    for (r = q->output_head; r; r = rn) {
-        //fprintf(stderr, "Discard output %d\n", r->serial);
-        rn = r->next;
-        hts_tpool_delete_result(r, free_results);
-    }
+    // Remove any queued output, thus ensuring we have room to flush.
+    r_head = q->output_head;
     q->output_head = q->output_tail = NULL;
     q->n_output = 0;
     pthread_mutex_unlock(&q->p->pool_m);
 
+    // Release memory.  This can be done unlocked now the lists have been
+    // removed from the queue
+    for (j = j_head; j; j = jn) {
+        //fprintf(stderr, "Discard input %d\n", j->serial);
+        jn = j->next;
+        if (j->job_cleanup) j->job_cleanup(j->arg);
+        free(j);
+    }
+
+    for (r = r_head; r; r = rn) {
+        //fprintf(stderr, "Discard output %d\n", r->serial);
+        rn = r->next;
+        if (r->result_cleanup) {
+            r->result_cleanup(r->data);
+            r->data = NULL;
+        }
+        hts_tpool_delete_result(r, free_results);
+    }
+
     // Wait for any jobs being processed to complete.
     // (TODO: consider how to cancel any currently processing jobs.
     // Probably this is too hard.)
     if (hts_tpool_process_flush(q) != 0)
         return -1;
 
-    // Discard any new output.
+    // Remove any new output.
     pthread_mutex_lock(&q->p->pool_m);
-    for (r = q->output_head; r; r = rn) {
-        //fprintf(stderr, "Discard output %d\n", r->serial);
-        rn = r->next;
-        hts_tpool_delete_result(r, free_results);
-    }
+    r_head = q->output_head;
     q->output_head = q->output_tail = NULL;
     q->n_output = 0;
 
@@ -906,6 +956,17 @@ int hts_tpool_process_reset(hts_tpool_process *q, int free_results) {
     pthread_cond_signal(&q->input_not_full_c);
     pthread_mutex_unlock(&q->p->pool_m);
 
+    // Discard unwanted output
+    for (r = r_head; r; r = rn) {
+        //fprintf(stderr, "Discard output %d\n", r->serial);
+        rn = r->next;
+        if (r->result_cleanup) {
+            r->result_cleanup(r->data);
+            r->data = NULL;
+        }
+        hts_tpool_delete_result(r, free_results);
+    }
+
     return 0;
 }
 
